import datetime
import functools
import importlib.util
import logging
import re

# from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Match, Tuple, Type

from requests.models import Response

from postgrestutils.typing import (
    DateOrStr,
    DatetimeOrStr,
    DatetimeParser,
    JsonDict,
    Parser,
    ParserMapping,
)

_DJANGO = importlib.util.find_spec("django") is not None
if _DJANGO:
    from ._django_utils import autofetch, try_django_parse_dt  # noqa: 401


logger = logging.getLogger("postgrestutils")

# this regex matches postgres' JSON formatting for timestamps
JSON_TIMESTAMP_REGEX = re.compile(
    r"(?P<year>\d{4})-"
    r"(?P<month>\d{2})-"
    r"(?P<day>\d{2})"
    r"T(?P<hour>\d{2}):"
    r"(?P<minute>\d{2}):"
    r"(?P<second>\d{2})\."
    r"(?P<microsecond>\d{1,6})"
    r"((?P<offsetsign>[+-])"
    r"(?P<offsethours>\d{2}):"
    r"(?P<offsetminutes>\d{2}))?$"
)


def _split_datetime_parts(
    match: Match[str],
) -> Tuple[Dict[str, int], Dict[str, str]]:
    datetime: Dict[str, int] = dict()
    timezone: Dict[str, str] = dict()
    for key, value in match.groupdict().items():
        if value:
            if "offset" not in key:
                datetime[key] = int(value)
            else:
                timezone[key] = value
    return datetime, timezone


def _try_python_parse_dt(value: str) -> DatetimeOrStr:
    """
    Attempt to parse value as a datetime using only python utilities.
    :param value: the string to parse
    :return: the parsed `datetime` or the original string if parsing failed
    """
    match = JSON_TIMESTAMP_REGEX.match(value)
    if match:
        dt, tz = _split_datetime_parts(match)
        if tz:
            sign = -1 if tz.get("offsetsign", "+") == "-" else 1
            tz = datetime.timezone(
                offset=sign
                * datetime.timedelta(
                    hours=int(tz.pop("offsethours")),
                    minutes=int(tz.pop("offsetminutes")),
                )
            )
            parsed_dt = datetime.datetime(**dt).replace(tzinfo=tz).astimezone()
        else:
            # naive datetime so we assume local time
            local_tz = datetime.datetime.now().astimezone().tzinfo
            parsed_dt = datetime.datetime(**dt).replace(tzinfo=local_tz)
        return parsed_dt
    return value


def _lazy_try_parse_dt() -> DatetimeParser:
    if _DJANGO:
        return try_django_parse_dt
    else:
        return _try_python_parse_dt


_try_parse_dt = _lazy_try_parse_dt()


def _try_parse_date(value: str) -> DateOrStr:
    try:
        return datetime.datetime.strptime(value, "%Y-%m-%d").date()
    except ValueError:
        return value


def _json_object_hook(
    json_dict: JsonDict,
    parsers_dict: Dict[Type[Any], Tuple[ParserMapping, ...]] = dict(),
    parser_flags: Parser = Parser.NONE,
) -> JsonDict:
    def _try_applying_parsers(
        json_dict: JsonDict,
        type_to_parse: Any,
        parser_mappings: Tuple[ParserMapping, ...],
    ):
        for parser, func in parser_mappings:
            if parser & parser_flags:
                maybe_parsed_value = func(value)

                # check if parsing succeeded
                if not isinstance(maybe_parsed_value, type_to_parse):
                    json_dict[key] = maybe_parsed_value
                    break

    for key, value in json_dict.items():
        type_to_parse = type(value)
        parsers = parsers_dict.get(type_to_parse)

        if parsers is not None:
            _try_applying_parsers(json_dict, type_to_parse, parsers)
    return json_dict


def parse_with_custom_parsers(resp: Response, parser: Parser):
    parsers_dict = {
        str: ((Parser.DATE, _try_parse_date), (Parser.DATETIME, _try_parse_dt)),
    }
    object_hook = functools.partial(
        _json_object_hook, parsers_dict=parsers_dict, parser_flags=parser
    )

    return resp.json(object_hook=object_hook)
